/*
 * tkPictImgFmtPPM.c --
 *
 *	A pict image file handler for PPM (Portable PixMap) files.
 *
 * Copyright (c) 1995 The Regents of the University of California.
 *
 * Author: Pierre-Louis Bossart
 * Date: November 17, 1995
 *
 * Derived from tkImgFmtPPM.c in the tk4.0b2 distribution 
 * copyrighted as follows:
 *
 *    Copyright (c) 1994 The Australian National University.
 *    Copyright (c) 1994-1995 Sun Microsystems, Inc.
 *
 *    Author: Paul Mackerras (paulus@cs.anu.edu.au),
 *	      Department of Computer Science,
 *	      Australian National University.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "tkpict.h"

/*
 * Define PGM and PPM, i.e. gray images and color images.
 */

#define PGM 1
#define PPM 2

/*
 * The format record for the PPM file format:
 */

static int		FileMatchPPM _ANSI_ARGS_((FILE *f, char *fileName,
			    char *formatString, int *widthPtr,
			    int *heightPtr));
static int		FileReadPPM  _ANSI_ARGS_((Tcl_Interp *interp,
			    FILE *f, char *fileName, char *formatString,
			    Tk_PictHandle imageHandle, int destX, int destY,
			    int width, int height, int srcX, int srcY));
static int		FileWritePPM _ANSI_ARGS_((Tcl_Interp *interp,
			    char *fileName, char *formatString,
			    Tk_PictImageBlock *blockPtr));

Tk_PictImageFormat tkImgFmtPPM = {
    "PPM",			/* name */
    FileMatchPPM,		/* fileMatchProc */
    NULL,			/* stringMatchProc */
    FileReadPPM,		/* fileReadProc */
    NULL,			/* stringReadProc */
    FileWritePPM,		/* fileWriteProc */
    NULL,			/* stringWriteProc */
};

/*
 * Prototypes for local procedures defined in this file:
 */

static int		ReadPPMFileHeader _ANSI_ARGS_((FILE *f, int *widthPtr,
			    int *heightPtr));

/*
 *----------------------------------------------------------------------
 *
 * FileMatchPPM --
 *
 *	This procedure is invoked by the pict image type to see if
 *	a file contains image data in PPM format.
 *
 * Results:
 *	The return value is >0 if the first characters in file "f" look
 *	like PPM data, and 0 otherwise.
 *
 * Side effects:
 *	The access position in f may change.
 *
 *----------------------------------------------------------------------
 */

static int
FileMatchPPM(f, fileName, formatString, widthPtr, heightPtr)
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here if the file is a valid
				 * raw PPM file. */
{
    return ReadPPMFileHeader(f, widthPtr, heightPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * FileReadPPM --
 *
 *	This procedure is called by the pict image type to read
 *	PPM format data from a file and write it into a given
 *	pict image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	The access position in file f is changed, and new data is
 *	added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

static int
FileReadPPM(interp, f, fileName, formatString, imageHandle, destX, destY,
	width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PictHandle imageHandle;	/* The pict image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * pict image to be written to. */
    int width, height;		/* Dimensions of block of pict image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
    int fileWidth, fileHeight;
    int nBytes, type, count;
    unsigned char *pixelPtr;
    Tk_PictImageBlock block;

    unsigned char *srcPtr;
    unsigned char *destPtr;
    unsigned char *newPixelPtr;
    int i;
   
   

    type = ReadPPMFileHeader(f, &fileWidth, &fileHeight);
    if (type == 0) {
	Tcl_AppendResult(interp, "couldn't read raw PPM header from file \"",
		fileName, "\"", NULL);
	return TCL_ERROR;
    }
    if ((fileWidth <= 0) || (fileHeight <= 0)) {
	Tcl_AppendResult(interp, "PPM image file \"", fileName,
		"\" has dimension(s) <= 0", (char *) NULL);
	return TCL_ERROR;
    }

    if ((srcX + width) > fileWidth) {
	width = fileWidth - srcX;
    }
    if ((srcY + height) > fileHeight) {
	height = fileHeight - srcY;
    }
    if ((width <= 0) || (height <= 0)
	|| (srcX >= fileWidth) || (srcY >= fileHeight)) {
	return TCL_OK;
    }

    if (type == PGM) {
        block.pixelSize = 1;
    }
    else {
        block.pixelSize = 3;
    }
    block.width = width;
    block.height = height;
    block.datatype = BYTE;
    block.pitch = block.pixelSize * fileWidth;
    block.skip = 0;

    if ( destX != 0 || destY != 0 || width != fileWidth )
      block.copy = COPY;
    else
      block.copy = NO_COPY; 

    Tk_PictExpand(imageHandle, destX + width, destY + height);

    if (srcY > 0) {
	fseek(f, (long) (srcY * block.pitch), SEEK_CUR);
    }

 
    nBytes = height* block.pitch;
    pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
    if(pixelPtr == NULL) {
      (void)fprintf(stderr,"FileReadPPM: Could not allocate memory\n");
      return;
    }

    block.pixelPtr = pixelPtr + srcX * block.pixelSize;

    count = fread(pixelPtr, 1, (unsigned) nBytes, f);
    if (count != nBytes) {
      Tcl_AppendResult(interp, "error reading PPM image file \"",
		       fileName, "\": ",
		       feof(f) ? "not enough data" : Tcl_PosixError(interp),
		       (char *) NULL);
      ckfree((char *) pixelPtr);
      return TCL_ERROR;
    }
	
    if  (type == PPM ) {
      block.pixelSize = 1;
      block.pitch = fileWidth;
      nBytes = height* block.pitch;

      newPixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
      if(newPixelPtr == NULL) {
	(void)fprintf(stderr,"FileReadPPM: Could not allocate memory\n");
	return;
      }
      srcPtr = pixelPtr;
      destPtr = newPixelPtr;

      for(i=0;i<nBytes;i++) {
	*destPtr = *srcPtr;
	destPtr++;
	srcPtr += 3;
      }
      free((void*)pixelPtr);
      block.pixelPtr = newPixelPtr + srcX * block.pixelSize;
    }
      
    Tk_PictPutBlock(imageHandle, &block, destX, destY, width, height);
	
    /* free data if necessary */
    if( block.copy == COPY )
      free((void*)pixelPtr);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * FileWritePPM --
 *
 *	This procedure is invoked to write image data to a file in PPM
 *	format.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	Data is written to the file given by "fileName".
 *
 *----------------------------------------------------------------------
 */

static int
FileWritePPM(interp, fileName, formatString, blockPtr)
    Tcl_Interp *interp;
    char *fileName;
    char *formatString;
    Tk_PictImageBlock *blockPtr;
{
    FILE *f;
    int w, h;
    int nBytes;
    unsigned char *pixelPtr, *pixLinePtr;

    if ((f = fopen(fileName, "w")) == NULL) {
      Tcl_AppendResult(interp, fileName, ": ", Tcl_PosixError(interp),
		       (char *)NULL);
      return TCL_ERROR;
    }
    if( blockPtr->datatype != BYTE ) {
      fprintf(f,"Convert first the image to type byte \n");
      return TCL_ERROR;
    }

    fprintf(f, "P5\n%d %d\n255\n", blockPtr->width, blockPtr->height);

    pixLinePtr = blockPtr->pixelPtr;
    nBytes = blockPtr->height * blockPtr->pitch;
    if (fwrite(pixLinePtr, 1, (unsigned) nBytes, f) != nBytes) {
      Tcl_AppendResult(interp, "error writing \"", fileName, "\": ",
		       Tcl_PosixError(interp), (char *) NULL);
      if (f != NULL) {
	fclose(f);
      }
      return TCL_ERROR;
    }
    if (fclose(f) == 0) {
      return TCL_OK;
    }
    f = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * ReadPPMFileHeader --
 *
 *	This procedure reads the PPM header from the beginning of a
 *	PPM file and returns the dimensions of the image.
 *
 * Results:
 *	The return value is PGM if file "f" appears to start with
 *	a valid PGM header, PPM if "f" appears to start with a valid
 *      PPM header, and 0 otherwise.  If the header is valid,
 *	then *widthPtr and *heightPtr are modified to hold the
 *	dimensions of the image.
 *
 * Side effects:
 *	The access position in f advances.
 *
 *----------------------------------------------------------------------
 */

static int
ReadPPMFileHeader(f, widthPtr, heightPtr)
    FILE *f;			/* Image file to read the header from */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here. */
{
    int maxPixel;
#define BUFFER_SIZE 1000
    char buffer[BUFFER_SIZE];
    int i, numFields, firstInLine, c;
    int type = 0;

    /*
     * Read 4 space-separated fields from the file, ignoring
     * comments (any line that starts with "#").
     */

    c = getc(f);
    firstInLine = 1;
    i = 0;
    for (numFields = 0; numFields < 4; numFields++) {
	/*
	 * Skip comments and white space.
	 */

	while (1) {
	    while (isspace(UCHAR(c))) {
		firstInLine = (c == '\n');
		c = getc(f);
	    }
	    if (c != '#') {
		break;
	    }
	    do {
		c = getc(f);
	    } while ((c != EOF) && (c != '\n'));
	    firstInLine = 1;
	}

	/*
	 * Read a field (everything up to the next white space).
	 */

	while ((c != EOF) && !isspace(UCHAR(c))) {
	    if (i < (BUFFER_SIZE-2)) {
		buffer[i]  = c;
		i++;
	    }
	    c = getc(f);
	}
	buffer[i] = ' ';
	i++;
	firstInLine = 0;
    }
    buffer[i] = 0;

    /*
     * Parse the fields, which are: id, width, height, maxPixel.
     */

    if (strncmp(buffer, "P6 ", 3) == 0) {
	type = PPM;
    } else if (strncmp(buffer, "P5 ", 3) == 0) {
	type = PGM;
    } else {
	return 0;
    }
    if (sscanf(buffer+3, "%d %d %d", widthPtr, heightPtr, &maxPixel) != 3) {
	return 0;
    }
    if (maxPixel != 255) {
	return 0;
    }

    slice_nb = 0;
    nb_slices = 1;

    return type;
}
